package Frame;
require CMONOS::CGI_Lib::Common;
@ISA = qw(Common);
use strict;
use utf8;
use Fcntl qw(:flock);
######################################################################
# 矩形モジュール
# Frame.pm 1.1.5
# Copyright (C) 1999-2018 CMONOS Co. Ltd.
#
# support@cmonos.jp
# http://cmonos.jp/
#
# 注意：CGI_Lib/Image.pm が必要です
# 最終更新日：2008-03-24
######################################################################
#
# このサブルーチン集の使い方
# ※ここでいうフレームとは（角丸）矩形のことです。
# ---------------------------------------------------------------------------------------
#	(0) オブジェクトを作る。
#		require CMONOS::CGI_Lib::Frame;
#		obj = Frame->new(%setting);
#
#	(0)-1 必要項目を設定する（パスは全て相対パスで指定して下さい）。
#		obj->set(
#			'###' => '',
#			'***･･･' => ''
#		)
#
#		### フレーム全体の設定
#			type				frame	標準型フレーム
#								tab	タブ型フレーム
#								bar	バー型フレーム
#			dir					フレーム素材ディレクトリのパス（settingで指定することもできます）
#			setting				フレーム設定のパス（dirにあるときは自動認識します）
#			spacer				透明GIFのパス（dirにあるときは自動認識します）
#			imagesize			"no"=角丸画像のサイズを取得しない
#			output				タグを書きだすファイルのパス
#			fixed				yes/no=フレームの内容をすべて読み込んでから表示する
#
#		*** 設定対象
#			frame				フレーム全体
#			tl					上段/左端
#			tc					上段/中央
#			tr					上段/右端
#			ml					中段/左端
#			mc					中段/中央
#			mr					中段/右端
#			bl					下段/左端
#			bc					下段/中央
#			br					下段/左端
#			titlebar			タイトル部
#			contents			フレームで囲む内容
#
#		･･･ 設定項目
#			image				画像のパス
#			background			背景画像のパス
#			bgcolor				背景色
#			align				水平方向位置
#			valign				垂直方向位置
#			width				横幅
#			height				高さ
#			class				クラス
#			id					ID
#
#		･･･ titlebar/contentsにのみ有効な設定項目
#			fontcolor			フォント色
#			fontsize			フォントサイズ
#			fontoption			フォントオプション
#			bgcolor				タイトルの背景色
#			cellpadding			タイトルと枠の間隔
#			nowrap				nowrap="nowrap"
#
#		例）	'framewidth' => '500',
#				'framealign' => 'center',
#				'tlimage' => './image.gif'
#
#	(0)-6 パスをoutputから見たパスに変換する（元に戻すときはoutputを空にして実行します。）
#		obj->frame_base_change(output,output_option)
#
# ---------------------------------------------------------------------------------------
#	(1) フレーム作成
#		obj->frame(\contents[,\titlebar[,option[,setting]]]);
# 			contents			フレームで囲む内容のリファレンス
# 			titlebar				フレームタイトルのリファレンス
# 			option			設定情報オプション
# 								all	 全ての設定情報をタグに含める
# 								dir	 素材ディレクトリ情報のみ
# 								tag	 タグのみ
# 								no		 タグを付けない
# 			setting			設定情報
#
# ---------------------------------------------------------------------------------------
#	(2) 背景付きテキスト作成
#		obj->box(contents,type);
# 			contents			背景を付ける内容
# 			type				titlebarとcontentsのどちらの設定を使うか指定
#
# ---------------------------------------------------------------------------------------
#	(3) obj->change_color(dir,out,type,@setting);
#		ディレクトリの中にあるGIF画像の配色表を一括変換します。
# 			dir				オリジナルフレームのあるディレクトリのパス
# 			out				変換したフレームを収めるディレクトリのパス
#
# 			※	この他の設定はImage.pm(6)-1/2に準じます。
#
# ---------------------------------------------------------------------------------------
#	(4)-1 obj->text_to_frame(text,option,\%obj,output);
#		テキスト中のフレーム表示指定タグ部分をフレーム表示に変換します。
# 			text				フレーム表示に変換するテキスト
# 			option			設定情報オプション
# 								all	 全ての設定情報をタグに含める
# 								dir	 素材ディレクトリ情報のみ
# 								tag	 タグのみ
# 								no		 タグを付けない
# 			obj				設定情報を持つオブジェクトを収めたハッシュ
# 								（key=オブジェクト名/value=オブジェクト）のリファレンス
# 								※ タグ中の id="オブジェクトキー" でオブジェクトを指定します。
# 			output			出力先パス
#
#		タグ
# 			<!--frame setting --><!--titlebar--><!--/titlebar-->
# 												<!--contents--><!--/contents--><!--/frame-->
# 			※ settingは、***･･･="設定値"
#
#	(4)-2 obj->frame_to_text(text);
#		テキスト中のフレーム表示指定タグ部分をテキストに戻します。
# 			text				フレーム表示を元に戻すテキスト
#
######################################################################
#オブジェクト作成
sub new ($;@) {
    my ( $pkg, @setting ) = @_;
    my $this = {};
    bless $this;
    unless ( ref $cgi::Image ) {
        require CMONOS::CGI_Lib::Image;
        $cgi::Image = Image->new;
    }
    $this->{'library'} = 'frame';
    $this->set(@setting) if @setting;
    $this;
}

# --------------------------------------------------------------------
#	(0)-1-0 初期設定
sub setting ($@) {
    my ( $this, @setting ) = @_;
    if ( @setting == 1 ) {
        my $setting = $setting[0];
        @setting = ();
        while ( $setting
            =~ /[\x09\x0A\x0C\x0D\x20]*([0-9a-zA-Z_]+)[\x09\x0A\x0C\x0D\x20]*=[\x09\x0A\x0C\x0D\x20]*"([^"]*)"/sg
            )
        {
            my ( $k, $v ) = ( $1, $2 );
            $v =~ s/&lt;/</sg;
            $v =~ s/&gt;/>/sg;
            $v =~ s/&quot;/"/sg;
            $v =~ s/&#10;/\x0A/sg;
            $v =~ s/&amp;/\&/sg;
            push( @setting, $k, $v );
        }
    }
    return @setting;
}

#	(0)-1 初期設定
sub set ($@) {
    my $this = shift;
    my ( $k, $v, %setting, @images );

    #文字列設定の処理
    %setting = $this->setting(@_);

    #設定ファイルがある場合は読み込む
    if (   defined $setting{'setting'}
        && $setting{'setting'} ne ''
        && -f $setting{'setting'} )
    {
        if ( open( PREF, $setting{'setting'} ) ) {
            binmode PREF, ":utf8";
            flock( PREF, LOCK_SH );
            $v = join( '', <PREF> );
            close(PREF);
            %setting = ( %setting, $this->setting($v) );
        }
    }

    #素材ディレクトリのスキャン
    if (   defined $setting{'dir'}
        && $setting{'dir'} ne ''
        && -d $setting{'dir'} )
    {
        $setting{'dir'} =~ s!/$!!;
        opendir( DIR, $setting{'dir'} );
        @images = grep( -f "$setting{'dir'}/$_" && $_ =~ /\.(gif|jpg|png)$/i,
            readdir(DIR) );
        closedir(DIR);

        foreach $k (
            'tl', 'tc',       'tr',       'ml',
            'mc', 'mr',       'bl',       'bc',
            'br', 'titlebar', 'contents', 'frame'
            )
        {
            ( $setting{ $k . 'image' } ) = grep( $_ =~ /^$k\./i, @images );
            ( $setting{ $k . 'background' } )
                = grep( $_ =~ /^${k}bg\./i, @images );

            $setting{ $k . 'image' }
                = $setting{'dir'} . '/' . $setting{ $k . 'image' }
                if defined $setting{ $k . 'image' }
                && $setting{ $k . 'image' } ne '';
            $setting{ $k . 'background' }
                = $setting{'dir'} . '/' . $setting{ $k . 'background' }
                if defined $setting{ $k . 'background' }
                && $setting{ $k . 'background' } ne '';
        }

        #空白GIF
        my ($spacer);
        ($spacer) = grep( $_ =~ /^spacer\.gif$/i, @images );
        $setting{'spacer'} = $setting{'dir'} . '/' . $spacer
            if ( !defined $setting{'spacer'} || $setting{'spacer'} eq '' )
            && defined $spacer
            && $spacer ne '';

        if ( open( PREF, "$setting{'dir'}/frame.pref" ) ) {
            binmode PREF, ":utf8";
            flock( PREF, LOCK_SH );
            $v = join( '', <PREF> );
            close(PREF);
            %setting = ( %setting, $this->setting($v) );
        }
    }
    while ( ( $k, $v ) = each %setting ) {
        next if $k eq '';
        $v = '' if !defined $v;
        $this->{$k} = $v;

        #画像サイズを調べる
        if ( !$this->{'cssbase'} && !$setting{'cssbase'} ) {
            if ((   $k =~ /image$/
                    || ( $k =~ /^(titlebar|contents)/ && $k =~ /background$/ )
                )
                && (   !defined $setting{ $k . 'width' }
                    || $setting{ $k . 'width' } eq ''
                    || !defined $setting{ $k . 'height' }
                    || $setting{ $k . 'height' } eq '' )
                )
            {
                ( $this->{ $k . 'width' }, $this->{ $k . 'height' } )
                    = $cgi::Image->info( $setting{$k} );
                if ( $k =~ /^([tb][rl])/ ) {
                    my $key = $1;
                    $this->{ $key . 'width' } = $setting{ $k . 'width' }
                        if !defined $setting{ $key . 'width' }
                        || $setting{ $key . 'width' } eq '';
                    $this->{ $key . 'height' } = $setting{ $k . 'height' }
                        if !defined $setting{ $key . 'height' }
                        || $setting{ $key . 'height' } eq '';
                }
            }
        }
    }
    if ( $this->{'cssbase'} ) {
        $this->cell_style;
    }
    else {
        $this->{'spacer'} = $env::spacer
            unless defined $this->{'spacer'}
            && -f $this->{'spacer'};    #空白GIF
        $this->cell_size( $setting{'mcwidth'}, $setting{'mcheight'} );
        $this->cell_align;
    }
    return $this;
}

#セルのスタイル
sub cell_style ($) {
    my $this = shift;
    my ( $k, $style );
    foreach $k (
        'frame', 'tl', 'tc',       'tr',
        'ml',    'mc', 'mr',       'bl',
        'bc',    'br', 'titlebar', 'contents'
        )
    {
        my $style = $k . 'style';
        $this->{$style} = '' if !defined $this->{$style};
        $this->replace_cell_style(
            \$this->{$style}, 'padding',
            ' padding: ' . $this->{ $k . 'cellpadding' } . 'px;',
            \$this->{ $k . 'cellpadding' }
            )
            if defined $this->{ $k . 'cellpadding' }
            && $this->{ $k . 'cellpadding' };
        $this->replace_cell_style(
            \$this->{$style}, 'width',
            ' width: ' . $this->{ $k . 'width' } . 'px;',
            \$this->{ $k . 'width' }
        ) if defined $this->{ $k . 'width' } && $this->{ $k . 'width' };
        $this->replace_cell_style(
            \$this->{$style}, 'height',
            ' height: ' . $this->{ $k . 'height' } . 'px;',
            \$this->{ $k . 'height' }
        ) if defined $this->{ $k . 'height' } && $this->{ $k . 'height' };
        if ( defined $this->{ $k . 'align' } && $this->{ $k . 'align' } ) {
            $this->replace_cell_style( \$this->{$style}, 'text-align',
                ' text-align: ' . $this->{ $k . 'align' } . ';' );
            if ( $k eq 'titlebar' || $k eq 'contents' ) {
                $this->replace_cell_style( \$this->{$style}, 'margin-left',
                    ' margin-left: auto;' )
                    if $this->{ $k . 'align' } eq 'center'
                    || $this->{ $k . 'align' } eq 'right';
                $this->replace_cell_style( \$this->{$style}, 'margin-right',
                    ' margin-right: auto;' )
                    if $this->{ $k . 'align' } eq 'center'
                    || $this->{ $k . 'align' } eq 'left';
            }
            delete $this->{ $k . 'align' };
        }
        if ( $k eq 'tc' && $this->{'titlebarvalign'} ) {
            $this->replace_cell_style(
                \$this->{$style},
                'vertical-align',
                ' vertical-align: ' . $this->{'titlebarvalign'} . ';',
                \$this->{'tcvalign'},
                \$this->{'titlebarvalign'}
            );
        }
        elsif ( $k eq 'mc' && $this->{'contentsvalign'} ) {
            $this->replace_cell_style(
                \$this->{$style},
                'vertical-align',
                ' vertical-align: ' . $this->{'contentsvalign'} . ';',
                \$this->{'mcvalign'},
                \$this->{'contentsvalign'}
            );
        }
        elsif ( defined $this->{ $k . 'valign' } && $this->{ $k . 'valign' } )
        {
            $this->replace_cell_style(
                \$this->{$style}, 'vertical-align',
                ' vertical-align: ' . $this->{ $k . 'valign' } . ';',
                \$this->{ $k . 'valign' }
            );
        }
        $this->replace_cell_style(
            \$this->{$style}, 'background-color',
            ' background-color: ' . $this->{ $k . 'bgcolor' } . ';',
            \$this->{ $k . 'bgcolor' }
            )
            if defined $this->{ $k . 'bgcolor' } && $this->{ $k . 'bgcolor' };
        $this->replace_cell_style(
            \$this->{$style},
            'background-image',
            ' background-image: url(&quot;'
                . $this->{ $k . 'background' }
                . '&quot;);',
            \$this->{ $k . 'background' }
            )
            if defined $this->{ $k . 'background' }
            && $this->{ $k . 'background' };
        $this->replace_cell_style(
            \$this->{$style}, 'white-space',
            ' white-space: nowrap;',
            \$this->{ $k . 'nowrap' }
        ) if defined $this->{ $k . 'nowrap' } && $this->{ $k . 'nowrap' };
        $this->replace_cell_style(
            \$this->{$style}, 'font-color',
            ' font-color: ' . $this->{ $k . 'fontcolor' } . ';',
            \$this->{ $k . 'fontcolor' }
            )
            if defined $this->{ $k . 'fontcolor' }
            && $this->{ $k . 'fontcolor' };
        $this->replace_cell_style(
            \$this->{$style},
            'font-size',
            ' font-size: '
                . $this->font_size_to_style( $this->{ $k . 'fontsize' } )
                . ';',
            \$this->{ $k . 'fontsize' }
            )
            if defined $this->{ $k . 'fontsize' }
            && $this->{ $k . 'fontsize' };
    }
}

sub replace_cell_style ($$$$$;$$) {
    my ( $this, $cell_style, $style_name, $style, $setting, $setting2 ) = @_;
    $style =~ s/[\x09\x0A\x0C\x0D\x20]+/ /sg;
    $$cell_style
        .= ( ( $$cell_style && $$cell_style !~ /;$/ ) ? ';' : '' ) . $style
        unless $$cell_style
        =~ s/[\x09\x0A\x0C\x0D\x20]*$style_name:.*?(;|$)/$style/gs;
    undef $$setting  if ref $setting;
    undef $$setting2 if ref $setting2;
}

sub font_size_to_style ($$) {
    my ( $this, $size ) = @_;
    if ( $size == 1 ) {
        return '0.67em';
    }
    elsif ( $size == 2 ) {
        return '0.83em';
    }
    elsif ( $size == 3 ) {
        return '1em';
    }
    elsif ( $size == 4 ) {
        return '1.17em';
    }
    elsif ( $size == 5 ) {
        return '1.5em';
    }
    elsif ( $size == 6 ) {
        return '2em';
    }
    elsif ( $size == 7 ) {
        return '3em';
    }
    else {
        return '1em';
    }
}

#セルのサイズ
sub cell_size ($$$) {
    my ( $this, $mcwidth, $mcheight ) = @_;

    $this->{'tlwidth'} = $this->{'tlimagewidth'}
        if ( !defined $this->{'tlwidth'} || $this->{'tlwidth'} eq '' )
        && defined $this->{'tlimagewidth'}
        && $this->{'tlimagewidth'} ne '';
    $this->{'tlheight'} = $this->{'tlimageheight'}
        if ( !defined $this->{'tlheight'} || $this->{'tlheight'} eq '' )
        && defined $this->{'tlimageheight'}
        && $this->{'tlimageheight'} ne '';
    $this->{'trwidth'} = $this->{'trimagewidth'}
        if ( !defined $this->{'trwidth'} || $this->{'trwidth'} eq '' )
        && defined $this->{'trimagewidth'}
        && $this->{'trimagewidth'} ne '';
    $this->{'trheight'} = $this->{'trimageheight'}
        if ( !defined $this->{'trheight'} || $this->{'trheight'} eq '' )
        && defined $this->{'trimageheight'}
        && $this->{'trimageheight'} ne '';

    $this->{'blwidth'} = $this->{'blimagewidth'}
        if ( !defined $this->{'blwidth'} || $this->{'blwidth'} eq '' )
        && defined $this->{'blimagewidth'}
        && $this->{'blimagewidth'} ne '';
    $this->{'blheight'} = $this->{'blimageheight'}
        if ( !defined $this->{'blheight'} || $this->{'blheight'} eq '' )
        && defined $this->{'blimageheight'}
        && $this->{'blimageheight'} ne '';
    $this->{'brwidth'} = $this->{'brimagewidth'}
        if ( !defined $this->{'brwidth'} || $this->{'brwidth'} eq '' )
        && defined $this->{'brimagewidth'}
        && $this->{'brimagewidth'} ne '';
    $this->{'brheight'} = $this->{'brimageheight'}
        if ( !defined $this->{'brheight'} || $this->{'brheight'} eq '' )
        && defined $this->{'brimageheight'}
        && $this->{'brimageheight'} ne '';

    if (   ( !defined $mcwidth || $mcwidth eq '' )
        && defined $this->{'framewidth'}
        && $this->{'framewidth'} =~ /^[0-9]+$/
        && $this->{'framewidth'} > 1
        && defined $this->{'tlwidth'}
        && $this->{'tlwidth'} =~ /^[0-9]+$/
        && $this->{'tlwidth'}
        && defined $this->{'trwidth'}
        && $this->{'trwidth'} =~ /^[0-9]+$/
        && $this->{'trwidth'} )
    {
        $this->{'mcwidth'}
            = $this->{'framewidth'} - $this->{'tlwidth'} - $this->{'trwidth'};
    }
    elsif ( defined $mcwidth && $mcwidth ) {
        $this->{'mcwidth'} = $mcwidth;
    }
    if (   ( !defined $mcheight || $mcheight eq '' )
        && defined $this->{'frameheight'}
        && $this->{'frameheight'} =~ /^[0-9]+$/
        && $this->{'frameheight'} > 1
        && defined $this->{'tlheight'}
        && $this->{'tlheight'} =~ /^[0-9]+$/
        && $this->{'tlheight'}
        && defined $this->{'blheight'}
        && $this->{'blheight'} =~ /^[0-9]+$/
        && $this->{'blheight'} )
    {
        $this->{'mcheight'} = $this->{'mlheight'} = $this->{'mrheight'}
            = $this->{'frameheight'}
            - $this->{'tlheight'}
            - $this->{'blheight'};
    }
    elsif ( defined $mcheight && $mcheight ) {
        $this->{'mcheight'} = $this->{'mlheight'} = $this->{'mrheight'}
            = $mcheight;
    }

    $this->{'mcheight'} = $this->{'mlheight'} = $this->{'mrheight'} = '100%'
        if defined $this->{'frameheight'}
        && $this->{'frameheight'} =~ /^[0-9]+\%$/;
    $this->{'mcwidth'} = '100%'
        if !defined $this->{'mcwidth'} || !$this->{'mcwidth'};
    $this->{'tcwidth'} = $this->{'bcwidth'} = $this->{'mcwidth'};

    my ($cell);
    $this->{'leftwidth'} = 0;
    foreach $cell ( 'tl', 'ml', 'bl' ) {
        $this->{'leftwidth'} = $this->{ $cell . 'width' }
            if $this->{ $cell . 'width' }
            && $this->{ $cell . 'width' } !~ /[^0-9]/
            && $this->{ $cell . 'width' } > $this->{'leftwidth'};
        $this->{'leftwidth'} = $this->{ $cell . 'imagewidth' }
            if $this->{ $cell . 'imagewidth' }
            && $this->{ $cell . 'imagewidth' } !~ /[^0-9]/
            && $this->{ $cell . 'imagewidth' } > $this->{'leftwidth'};
    }
    $this->{'rightwidth'} = 0;
    foreach $cell ( 'tr', 'mr', 'br' ) {
        $this->{'rightwidth'} = $this->{ $cell . 'width' }
            if $this->{ $cell . 'width' }
            && $this->{ $cell . 'width' } !~ /[^0-9]/
            && $this->{ $cell . 'width' } > $this->{'rightwidth'};
        $this->{'rightwidth'} = $this->{ $cell . 'imagewidth' }
            if $this->{ $cell . 'imagewidth' }
            && $this->{ $cell . 'imagewidth' } !~ /[^0-9]/
            && $this->{ $cell . 'imagewidth' } > $this->{'rightwidth'};
    }
    $this->{'topheight'} = 0;
    foreach $cell ( 'tl', 'tc', 'tr' ) {
        $this->{'topheight'} = $this->{ $cell . 'height' }
            if $this->{ $cell . 'height' }
            && $this->{ $cell . 'height' } !~ /[^0-9]/
            && $this->{ $cell . 'height' } > $this->{'topheight'};
        $this->{'topheight'} = $this->{ $cell . 'imageheight' }
            if $this->{ $cell . 'imageheight' }
            && $this->{ $cell . 'imageheight' } !~ /[^0-9]/
            && $this->{ $cell . 'imageheight' } > $this->{'topheight'};
    }
    $this->{'bottomheight'} = 0;
    foreach $cell ( 'bl', 'bc', 'br' ) {
        $this->{'bottomheight'} = $this->{ $cell . 'height' }
            if $this->{ $cell . 'height' }
            && $this->{ $cell . 'height' } !~ /[^0-9]/
            && $this->{ $cell . 'height' } > $this->{'bottomheight'};
        $this->{'bottomheight'} = $this->{ $cell . 'imageheight' }
            if $this->{ $cell . 'imageheight' }
            && $this->{ $cell . 'imageheight' } !~ /[^0-9]/
            && $this->{ $cell . 'imageheight' } > $this->{'bottomheight'};
    }
}

#セルの位置
sub cell_align ($) {
    my $this = shift;
    my ($k);
    foreach $k ( 'tl', 'tc', 'tr', 'ml', 'mc', 'mr', 'bl', 'bc', 'br' ) {

        #背景
        $this->{ $k . 'background' } = $this->{'framebackground'}
            if defined $this->{'framebackground'}
            && $this->{'framebackground'}
            && ( !defined $this->{ $k . 'background' }
            || $this->{ $k . 'background' } eq '' );
        $this->{ $k . 'bgcolor' } = $this->{'framebgcolor'}
            if defined $this->{'framebgcolor'}
            && $this->{'framebgcolor'}
            && ( !defined $this->{ $k . 'bgcolor' }
            || $this->{ $k . 'bgcolor' } eq '' );

        #位置情報
        if ( !defined $this->{ $k . 'align' }
            || $this->{ $k . 'align' } eq '' )
        {
            if ( $k =~ /l$/ ) {
                $this->{ $k . 'align' } = 'right';
            }
            elsif (
                $k eq 'bc'
                || ( ( !defined $this->{'type'} || $this->{'type'} ne 'bar' )
                    && $k eq 'tc' )
                )
            {
                $this->{ $k . 'align' } = 'center';
            }
            elsif ( $k =~ /r$/ ) {
                $this->{ $k . 'align' } = 'left';
            }
        }
        if ( !defined $this->{ $k . 'valign' }
            || $this->{ $k . 'valign' } eq '' )
        {
            if ( $k =~ /^t/ ) {
                $this->{ $k . 'valign' } = 'bottom';
            }
            elsif ( $k =~ /^m[rl]$/ ) {
                $this->{ $k . 'valign' } = 'middle';
            }
            elsif ( $k =~ /^b/ ) {
                $this->{ $k . 'valign' } = 'top';
            }
        }

        #サイズ情報
        $this->{ $k . 'width' } = '0%'
            if $k =~ /^[tb][lr]$/
            && ( !defined $this->{ $k . 'width' }
            || $this->{ $k . 'width' } eq '' );
        $this->{ $k . 'height' } = '0%'
            if $k =~ /^[tb][lr]$/
            && ( !defined $this->{ $k . 'height' }
            || $this->{ $k . 'height' } eq '' );
    }
}

# --------------------------------------------------------------------
#	(0)-2 設定の取得
sub get_setting ($;$$) {
    my ( $this, $key, $ng ) = @_;
    my ( $k, $v, $option );
    while ( ( $k, $v ) = each %{$this} ) {
        next
            if ref $v
            || !defined $v
            || $k eq 'Image'
            || $k eq 'setting'
            || $k eq 'dir'
            || $k eq 'output'
            || $k eq 'output_option'
            || $k eq ''
            || $v eq '';
        next if $key ne '' && $k !~ /$key/;
        next if $ng ne ''  && $k =~ /$ng/;
        $v =~ s/\&/&amp;/sg;
        $v =~ s/</&lt;/sg;
        $v =~ s/>/&gt;/sg;
        $v =~ s/"/&quot;/sg;
        $v =~ s/\x0A/&#10;/g;
        $option .= ' ' . $k . '="' . $v . '"';
    }
    return $option;
}

# --------------------------------------------------------------------
#	(0)-3 全ての設定の取得
sub get_all_setting ($) {
    my $this = shift;
    my ( $k, $v, @option );
    while ( ( $k, $v ) = each %{$this} ) {
        next
            if ref $v
            || !defined $v
            || $k eq 'Image'
            || $k eq 'setting'
            || $k eq 'dir'
            || $k eq 'output'
            || $k eq 'output_option'
            || $k eq ''
            || $v eq '';
        $v =~ s/\&/&amp;/sg;
        $v =~ s/</&lt;/sg;
        $v =~ s/>/&gt;/sg;
        $v =~ s/"/&quot;/sg;
        $v =~ s/\x0A/&#10;/g;
        push( @option, $k, $v );
    }
    return @option;
}

# --------------------------------------------------------------------
#	(0)-4 設定値の調整をせず直接設定
sub set_direct (@) {
    my ( $this, @setting ) = @_;
    @setting = $this->setting(@setting) if @setting == 1;
    my ( $k, $v, $framewidth, $frameheight, $mcwidth, $mcheight );
    while ( ( $k, $v ) = splice( @setting, 0, 2 ) ) {
        next if $k eq '';
        $this->{$k} = $v;
    }
    if ( $this->{'cssbase'} ) {
        $this->cell_style;
    }
    else {
        $this->cell_size;
        $this->cell_align;
    }
    return $this;
}

# --------------------------------------------------------------------
#	(0)-5 設定のリセット
sub reset_all ($) {
    my $this = shift;
    undef %{$this};
    return $this;
}

# --------------------------------------------------------------------
#	(0)-6 ベースURLの変更
sub frame_base_change ($;$$) {
    my ( $this, $output, $option ) = @_;
    my ( $k, $v );
    if (   defined $output
        && defined $this->{'output'}
        && $output eq $this->{'output'}
        && defined $option
        && defined $this->{'output_option'}
        && $option eq $this->{'output_option'} )
    {
        return $this;
    }
    elsif ( $output || $option ) {
        while ( ( $k, $v ) = each %{$this} ) {
            $this->{$k} = $this->base_change( $v, $output, $option )
                if $k =~ /(image|background|spacer)$/ && $v ne '';
        }
        $this->{'output'}        = $output if $output;
        $this->{'output_option'} = $option if $option;
    }
    elsif ( $this->{'output'} || $this->{'output_option'} ) {
        while ( ( $k, $v ) = each %{$this} ) {
            $this->{$k}
                = $this->un_base_change( $v, $this->{'output'},
                ( ( $this->{'output_option'} ) ? 'abs' : 'rel' ) )
                if $k =~ /(image|background|spacer)$/ && $v ne '';
        }
    }
    else {
        delete $this->{'output'};
        delete $this->{'output_option'};
    }
    return $this;
}

######################################################################
#	(1) obj->frame;
sub frame ($$;$$$$$) {
    my ( $this, $contents, $title, $option, $setting, $frame_class, $id )
        = @_;
    my ( $out, $contents_colspan, $title_rowspan, $temp, $class );
    if ( defined $this->{'type'} && $this->{'type'} eq 'tab' ) {
        $contents_colspan = 3;
    }
    elsif ( defined $this->{'type'} && $this->{'type'} eq 'bar' ) {
        $title_rowspan = 3;
        if (   defined $$contents
            && $$contents ne ''
            && defined $$title
            && $$title eq '' )
        {
            $title = $contents;
        }
        elsif ( defined $$contents && $$contents ne '' ) {
            $temp = ( ref $title && defined $$title ) ? $$title : '';
            $temp = '<h1 class="frame-title">' . $temp . '</h1>'
                if $temp ne '' && $temp !~ /<[^<>]*>/;
            $temp .= $$contents;
            $title = \$temp;
        }
    }
    elsif ( !defined $this->{'type'} || $this->{'type'} eq '' ) {
        $this->{'type'} = 'frame';
    }

    $class = $this->{'library'}
        . ( ( $this->{'id'} ) ? '-' . $this->{'id'} : '' );
    $out
        = $this->frame_header(
        ( ( ref $title && defined $$title && $$title ne '' ) ? 1 : 0 ),
        $class . ( ($frame_class) ? ' ' . $frame_class : '' ), $id );

    $out .= '<tr class="' . $class . '-top">';
    $out .= $this->cell( 'tl', $class, 'left' );
    $out .= $this->cell( 'tc', $class, 'center', 1, 'titlebar', $title, '',
        $title_rowspan );
    $out .= $this->cell( 'tr', $class, 'right' );
    $out .= '</tr>';

    $out .= '<tr class="' . $class . '-middle">';
    $out .= $this->cell( 'ml', $class, 'left', 1 )
        if $this->{'type'} ne 'tab';
    $out
        .= $this->cell( 'mc', $class, 'center', 0, 'contents', $contents,
        $contents_colspan )
        if $this->{'type'} ne 'bar';

    if ( $this->{'type'} ne 'tab' ) {
        $out .= $this->cell( 'mr', $class, 'right', 1 );
        $out .= '</tr>';

        $out .= '<tr class="' . $class . '-bottom">';
        $out .= $this->cell( 'bl', $class, 'left' );
        $out .= $this->cell( 'bc', $class, 'center', 1 )
            if $this->{'type'} ne 'bar';
        $out .= $this->cell( 'br', $class, 'right' );
    }
    $out .= '</tr>';

    $out .= '</tbody></table>';

    if ( defined $option && $option eq 'all' ) {
        $setting = $this->get_setting( '', '^(dir|setting|id)$' );
    }
    elsif ( defined $this->{'id'} && $this->{'id'} ne '' ) {
        $setting .= ' id="' . $this->{'id'} . '"';
        $setting .= ' type="' . $this->{'type'} . '"';
    }
    elsif ( defined $option && $option eq 'dir' ) {
        $setting = $this->get_setting('^(dir|type)$');
    }
    else {
        $setting .= ' type="' . $this->{'type'} . '"';
    }
    $out = '<!--_frame' . $setting . '-->' . $out . '<!--/_frame-->'
        if $option ne 'no';

    return $out;
}

# --------------------------------------------------------------------
#	(1)-0-1 テーブルタグ
sub frame_header ($;$$$) {
    my ( $this, $title_flag, $class, $id ) = @_;
    my ( $out, $style );
    $style = $this->escape( $this->{'framestyle'} )
        if defined $this->{'framestyle'} && $this->{'framestyle'};
    $out = '<table';
    if ($id) {
        $out .= ' id="' . $id . '"';
    }
    else {
        $out .= ' id="' . $this->{'frameid'} . '"'
            if defined $this->{'frameid'} && $this->{'frameid'};
    }
    $out
        .= ' class="'
        . ( ($class) ? $class : '' )
        . ( ( $this->{'frameclass'} ) ? ' ' . $this->{'frameclass'} : '' )
        . '"';
    $out .= ' cellspacing="0"';
    if ( !$this->{'cssbase'} ) {
        $out .= ' cellpadding="0" border="0"';
        $out .= ' align="' . $this->{'framealign'} . '"'
            if defined $this->{'framealign'} && $this->{'framealign'};
        $out
            .= ' width="'
            . ( ( defined $this->{'framewidth'} && $this->{'framewidth'} )
            ? $this->{'framewidth'}
            : '1' )
            . '"';
        $out .= ' height="'
            . (
            (          $this->{'type'} ne 'bar'
                    && defined $this->{'frameheight'}
                    && $this->{'frameheight'}
            ) ? $this->{'frameheight'} : '1'
            ) . '"';
        $out .= ' align="' . $this->{'framealign'} . '"'
            if defined $this->{'framealign'} && $this->{'framealign'};
        $out .= ' valign="' . $this->{'framevalign'} . '"'
            if defined $this->{'framevalign'} && $this->{'framevalign'};
        $out .= ' bgcolor="' . $this->{'framebgcolor'} . '"'
            if defined $this->{'framebgcolor'} && $this->{'framebgcolor'};
        $out .= ' background="' . $this->{'framebackground'} . '"'
            if defined $this->{'framebackground'}
            && $this->{'framebackground'};
        $style .= ' table-layout: fixed;'
            if defined $this->{'fixed'}
            && $this->{'fixed'} eq 'yes'
            && !$title_flag
            && !$this->{'contentsnowrap'}
            && defined $this->{'framewidth'}
            && $this->{'framewidth'} > 0;
        $out .= ' style="empty-cells: show;'
            . ( ($style) ? $style : '' ) . '"';
    }
    elsif ($style) {
        $out .= ' style="' . $style . '"';
    }
    $out .= '><tbody>';
    return $out;
}

# --------------------------------------------------------------------
#	(1)-0-2 Cell
sub cell ($$;$$$$$$) {
    my ($this,     $name,                 $frame_class,
        $td_class, $frame_object_enabled, $type,
        $contents, $colspan,              $rowspan
    ) = @_;
    my $out;
    $out = '<td';
    $out .= ' id="' . $this->{ $name . 'id' } . '"'
        if defined $this->{ $name . 'id' } && $this->{ $name . 'id' };
    $out
        .= ' class="'
        . ( ($frame_class) ? $frame_class . '-' : '' )
        . ( ($td_class)    ? $td_class          : '' )
        . ( ( $this->{ $name . 'class' } )
        ? ' ' . $this->{ $name . 'class' }
        : '' )
        . '"';
    $out .= ' colspan="' . $colspan . '"' if $colspan;
    $out .= ' rowspan="' . $rowspan . '"' if $rowspan;
    if ( !$this->{'cssbase'} ) {
        $out .= ' width="' . $this->{ $name . 'width' } . '"'
            if defined $this->{ $name . 'width' }
            && $this->{ $name . 'width' };
        $out .= ' height="' . $this->{ $name . 'height' } . '"'
            if defined $this->{ $name . 'height' }
            && $this->{ $name . 'height' };
        $out .= ' align="' . $this->{ $name . 'align' } . '"'
            if defined $this->{ $name . 'align' }
            && $this->{ $name . 'align' };
        if (   $type
            && defined $this->{ $type . 'valign' }
            && $this->{ $type . 'valign' } )
        {
            $out .= ' valign="' . $this->{ $type . 'valign' } . '"';
        }
        elsif ( defined $this->{ $name . 'valign' }
            && $this->{ $name . 'valign' } )
        {
            $out .= ' valign="' . $this->{ $name . 'valign' } . '"';
        }
        $out .= ' bgcolor="' . $this->{ $name . 'bgcolor' } . '"'
            if defined $this->{ $name . 'bgcolor' }
            && $this->{ $name . 'bgcolor' };
        $out .= ' background="' . $this->{ $name . 'background' } . '"'
            if defined $this->{ $name . 'background' }
            && $this->{ $name . 'background' };
        $out .= ' nowrap="nowrap"'
            if ( defined $this->{ $name . 'nowrap' }
            && $this->{ $name . 'nowrap' } )
            || ( defined $name && $name eq 'tc' );
    }
    $out .= ' style="' . $this->escape( $this->{ $name . 'style' } ) . '"'
        if defined $this->{ $name . 'style' } && $this->{ $name . 'style' };
    $out .= '>';

    if ( ref $contents && defined $$contents && $$contents ) {
        $out .= $this->box( $contents, $frame_class, $type, 'no' );
    }
    else {
        $out .= '<div class="' . $frame_class . '-object">'
            if $frame_object_enabled;
        if ( defined $this->{ $name . 'image' }
            && $this->{ $name . 'image' } )
        {
            if ( defined $this->{'imagesize'}
                && $this->{'imagesize'} eq 'no' )
            {
                $out .= $cgi::Image->image_tag(
                    $this->{ $name . 'image' }, '',
                    '',                         0,
                    '',                         ' style="display: block"',
                    '',                         '',
                    $this->{'output'},          $this->{'output_option'}
                );
            }
            else {
                $out .= $cgi::Image->image_tag(
                    $this->{ $name . 'image' },
                    $this->{ $name . 'imagewidth' },
                    $this->{ $name . 'imageheight' },
                    0,
                    '',
                    ' style="display: block"',
                    '',
                    '',
                    $this->{'output'},
                    $this->{'output_option'}
                );
            }
        }
        elsif ( !$this->{'cssbase'} ) {
            $out
                .= ( $this->{'spacer'} )
                ? '<img src="'
                . $this->{'spacer'}
                . '" border="0" width="1" height="1" style="display: block" />'
                : '<br />';
        }
        $out .= '</div>' if $frame_object_enabled;
    }
    $out .= '</td>';

    return $out;
}

# --------------------------------------------------------------------
#	(1)-0-3 Font
sub font ($$) {
    my ( $this, $contents, $type ) = @_;
    my $out;
    $out .= ' size="' . $this->{ $type . 'fontsize' } . '"'
        if defined $this->{ $type . 'fontsize' }
        && $this->{ $type . 'fontsize' };
    $out .= ' color="' . $this->{ $type . 'fontcolor' } . '"'
        if defined $this->{ $type . 'fontcolor' }
        && $this->{ $type . 'fontcolor' };
    $out .= ' ' . $this->{ $type . 'fontoption' }
        if defined $this->{ $type . 'fontoption' }
        && $this->{ $type . 'fontoption' };
    return ( defined $out && $out )
        ? '<font' . $out . '>' . $$contents . '</font>'
        : $$contents;
}

######################################################################
#	(3) obj->box
sub box ($$$) {
    my ( $this, $contents, $frame_class, $type, $tag, $setting ) = @_;
    my ( $div_flag, $out );

    #タグ
    if ( !defined $tag || $tag ne 'no' ) {
        $out .= '<!--_frame type="box"';
        if ( defined $setting && $setting ne '' ) {
            $out .= ' ' . $setting;
        }
        elsif ( $tag ne 'tag' ) {
            if ( !$this->{'cssbase'} ) {
                $out .= ' ' . $type . 'id="' . $this->{ $type . 'id' } . '"'
                    if defined $this->{ $type . 'id' }
                    && $this->{ $type . 'id' };
                $out
                    .= ' '
                    . $type
                    . 'class="'
                    . $this->{ $type . 'class' } . '"'
                    if defined $this->{ $type . 'class' }
                    && $this->{ $type . 'class' };
                $out
                    .= ' '
                    . $type
                    . 'cellpadding="'
                    . $this->{ $type . 'cellpadding' } . '"'
                    if defined $this->{ $type . 'cellpadding' }
                    && $this->{ $type . 'cellpadding' };
                $out
                    .= ' '
                    . $type
                    . 'backgroundheight="'
                    . $this->{ $type . 'backgroundheight' } . '"'
                    if defined $this->{ $type . 'backgroundheight' }
                    && $this->{ $type . 'backgroundheight' };
                $out
                    .= ' '
                    . $type
                    . 'backgroundwidth="'
                    . $this->{ $type . 'backgroundwidth' } . '"'
                    if defined $this->{ $type . 'backgroundwidth' }
                    && $this->{ $type . 'backgroundwidth' };
                $out
                    .= ' '
                    . $type
                    . 'align="'
                    . $this->{ $type . 'align' } . '"'
                    if defined $this->{ $type . 'align' }
                    && $this->{ $type . 'align' };
                $out
                    .= ' '
                    . $type
                    . 'valign="'
                    . $this->{ $type . 'valign' } . '"'
                    if defined $this->{ $type . 'valign' }
                    && $this->{ $type . 'valign' };
                $out
                    .= ' '
                    . $type
                    . 'bgcolor="'
                    . $this->{ $type . 'bgcolor' } . '"'
                    if defined $this->{ $type . 'bgcolor' }
                    && $this->{ $type . 'bgcolor' };
                $out
                    .= ' '
                    . $type
                    . 'background="'
                    . $this->{ $type . 'background' } . '"'
                    if defined $this->{ $type . 'background' }
                    && $this->{ $type . 'background' };
                $out .= ' ' . $type . 'nowrap="nowrap"'
                    if defined $this->{ $type . 'nowrap' }
                    && $this->{ $type . 'nowrap' };
            }
            $out
                .= ' '
                . $type
                . 'style="'
                . $this->escape( $this->{ $type . 'style' } ) . '"'
                if defined $this->{ $type . 'style' }
                && $this->{ $type . 'style' };
        }
        $out .= '-->';
    }

    #ボックス
    $out .= '<div';
    $out .= ' id="' . $this->{ $type . 'id' } . '"'
        if defined $this->{ $type . 'id' } && $this->{ $type . 'id' };
    $out
        .= ' class="'
        . ( ($frame_class) ? $frame_class . '-' : '' )
        . $type
        . ( ( $this->{ $type . 'class' } )
        ? ' ' . $this->{ $type . 'class' }
        : '' )
        . '"';
    if (   $this->{ $type . 'style' }
        || $this->{ $type . 'background' }
        || $this->{ $type . 'bgcolor' }
        || $this->{ $type . 'cellpadding' }
        || $this->{ $type . 'nowrap' }
        || $this->{ $type . 'align' }
        || $this->{ $type . 'valign' } )
    {
        $out .= ' style="';
        if ( !$this->{'cssbase'} ) {
            $out .= ' padding: ' . $this->{ $type . 'cellpadding' } . 'px;'
                if defined $this->{ $type . 'cellpadding' }
                && $this->{ $type . 'cellpadding' };
            $out
                .= ' height: '
                . $this->{ $type . 'backgroundheight' } . 'px;'
                if defined $this->{ $type . 'backgroundheight' }
                && $this->{ $type . 'backgroundheight' };
            $out .= ' width: ' . $this->{ $type . 'backgroundwidth' } . 'px;'
                if defined $this->{ $type . 'backgroundwidth' }
                && $this->{ $type . 'backgroundwidth' };
            $out .= ' text-align: ' . $this->{ $type . 'align' } . ';'
                if defined $this->{ $type . 'align' }
                && $this->{ $type . 'align' };
            $out .= ' margin-left: auto;'
                if $this->{ $type . 'align' } eq 'center'
                || $this->{ $type . 'align' } eq 'right';
            $out .= ' margin-right: auto;'
                if $this->{ $type . 'align' } eq 'center'
                || $this->{ $type . 'align' } eq 'left';
            $out
                .= ' background-color: ' . $this->{ $type . 'bgcolor' } . ';'
                if defined $this->{ $type . 'bgcolor' }
                && $this->{ $type . 'bgcolor' };
            $out
                .= ' background-image: url(&quot;'
                . $this->{ $type . 'background' }
                . '&quot;);'
                if defined $this->{ $type . 'background' }
                && $this->{ $type . 'background' };
            $out .= ' white-space: nowrap;'
                if defined $this->{ $type . 'nowrap' }
                && $this->{ $type . 'nowrap' };
        }
        $out .= ' ' . $this->escape( $this->{ $type . 'style' } )
            if defined $this->{ $type . 'style' }
            && $this->{ $type . 'style' };
        $out .= '"';
    }
    $out .= '>';

    $out .= '<!--' . $type . '-->';
    $out
        .= ( defined $$contents )
        ? ( $this->{'cssbase'} )
            ? $$contents
            : $this->font( $contents, $type )
        : '';
    $out .= '<!--/' . $type . '-->';
    $out .= '</div>';
    $out .= '<!--/_frame-->' if !defined $tag || $tag ne 'no';

    return $out;
}

######################################################################
#	(4)  ディレクトリの中にあるGIF画像を一括変換する
sub change_color ($$$$;@) {
    my ( $this, $dir, $out, $type, @setting ) = @_;

    $dir =~ s!/$!!;
    $out =~ s!/$!!;
    return unless defined $dir && defined $out && -d $dir && -d $out;

    if ( defined $setting[0]
        && ( $setting[0] eq 'slide' || $setting[0] eq 'change' ) )
    {
        $cgi::Image->change_color_table_multi( $dir, $out, $type, @setting );
        $dir .= '/frame.pref';
        $out .= '/frame.pref';

        unlink $out;
        $this->file_copy( $dir, $out );
        chmod( $env::file_mode, $out );
    }
    else {
        my (@files);
        opendir( DIR, $dir );
        @files = grep( -f "$dir/$_", readdir(DIR) );
        closedir(DIR);

        foreach (@files) {
            $this->file_copy( "$dir/$_", "$out/$_" );
        }
        chmod( $env::dir_mode, $out );
    }

}

######################################################################
#	(5)-1 obj->text_to_frame(\text,option,\obj,$output);
sub text_to_frame ($$;$$$$) {
    my ( $this, $text, $option, $obj, $output, $abs ) = @_;
    1 while $text
        =~ s/<!--frame([^>]*)-->((.(?!<!--frame[^>]*-->))*?)<!--\/frame-->/$this->_text_to_frame($1,$2,$option,$obj,$output,$abs)/egs;
    return $text;
}

sub _text_to_frame ($$$$;$$$) {
    my ( $this, $setting, $text, $option, $object, $output, $abs ) = @_;
    my ( $id, $contents, $title, $out, $type, $obj, $undef );
    $setting = '' if !defined $setting;
    $type    = $1 if $setting =~ s/[\x09\x0A\x0C\x0D\x20]*type="([^"]*)"//i;
    if ( $setting
        =~ s/[\x09\x0A\x0C\x0D\x20]*id[\x09\x0A\x0C\x0D\x20]*=[\x09\x0A\x0C\x0D\x20]*"([0-9]+)"//i
        )
    {
        $id = $1;
        if ( $setting =~ /=/ ) {
            $obj = Frame->new;
            if ( ref $$object{$id} ) {
                %$obj = %{ $$object{$id} };
            }
            else {
                %$obj = %$this;
            }
            $obj->set($setting);
            $undef = 1;
        }
        elsif ( ref $$object{$id} ) {
            $obj = $$object{$id};
        }
        else {
            $obj = $this;
        }
        $obj->{'id'} = $id;
    }
    elsif ( $setting =~ /\=/ ) {
        $obj   = Frame->new($setting);
        $undef = 1;
    }
    else {
        $obj = $this;
    }

    $obj->frame_base_change( $output, $abs );
    $contents = $1 if $text =~ s/<!--contents-->(.*)<!--\/contents-->//s;
    $title    = $1 if $text =~ s/<!--titlebar-->(.*)<!--\/titlebar-->//s;

    #ボックス
    if ( defined $type && $type eq "box" ) {
        return ( defined $contents && $contents ne '' )
            ? $obj->box( \$contents,
            'box' . ( ( $obj->{'id'} ) ? '-' . $obj->{'id'} : '' ),
            'contents', 'tag', $setting )
            : ( defined $title && $title ne '' ) ? $obj->box( \$title,
            'box' . ( ( $obj->{'id'} ) ? '-' . $obj->{'id'} : '' ),
            'titlebar', 'tag', $setting )
            : '';
    }

    $obj->{'type'} = $type if defined $type && $type;

    #バー型
    if ( defined $obj->{'type'} && $obj->{'type'} eq 'bar' ) {
        $title = $text if !defined $title || $title eq '';
    }
    else {
        $contents = $text if !defined $contents || $contents eq '';
    }
    $out = $obj->frame( \$contents, \$title, $option, $setting );

    undef %$obj if defined $undef && $undef;
    return $out;
}

# --------------------------------------------------------------------
#	(5)-2 obj->frame_to_text(\text);
sub frame_to_text ($$) {
    my ( $this, $text ) = @_;
    1 while $text
        =~ s/<!--_frame([^>]*)-->((.(?!<!--_frame[^>]*-->))*?)<!--\/_frame-->/$this->_frame_to_text($1,$2)/egs;
    return $text;
}

sub _frame_to_text ($$$$) {
    my ( $this, $setting, $text ) = @_;
    my ($out);
    $out .= $1 if $text =~ /(<!--titlebar-->.*<!--\/titlebar-->)/s;
    $out .= $1 if $text =~ /(<!--contents-->.*<!--\/contents-->)/s;
    $out = '<!--frame' . $setting . '-->' . $out . '<!--/frame-->';
    return $out;
}

1;
