package Session;
use Fcntl qw(:flock);
use strict;
use utf8;
######################################################################
# セッションモジュール
# Session.pm 1.9.2
# Copyright (C) 1999-2018 CMONOS Co. Ltd.
#
# support@cmonos.jp
# http://cmonos.jp/
# 最終更新日：2015-01-18
######################################################################
#
# このサブルーチン集の使い方
# ---------------------------------------------------------------------------------------
#	(0) オブジェクトを作る。
#		require CMONOS::CGI_Lib::Session;
#		obj = Session->new(store,security,lifetime);
#			store			セッション情報格納ディレクトリ
#			security		セッションハイジャック防止
#									ip=前回認証時と同一IPでない場合パスワードを再確認する
#									agent=前回認証時と同一エージェントでない場合パスワードを再確認する
#			lifetime		セッションの保存期間（日数）
#
# ---------------------------------------------------------------------------------------
#	(1) obj->is_available(id);
#		id が存在するかチェックします。存在すれば 1 を返します。
#
# ---------------------------------------------------------------------------------------
#	(2) (email,pswd) = obj->get_user_info(id);
#		id にユーザー認証情報が含まれていれば、それを返します。
#
# ---------------------------------------------------------------------------------------
#	(3) id = obj->mk_session_id;
#		セッションIDを作ります。
#
# ---------------------------------------------------------------------------------------
#	(4) id = obj->set_user_info(email,pswd,id);
#		ユーザー情報を記録します。
#
# ---------------------------------------------------------------------------------------
#	(5) obj->erase(id);
#		セッション情報を削除します。
#
# ---------------------------------------------------------------------------------------
#	(6) obj->gc;
#		保存期間を過ぎたセッションを削除します。
#
# ---------------------------------------------------------------------------------------
#	(7)-1 obj->save_by_uri(id,uri,\%data,read_flag,\%additional_values);
#		データを格納します。
#			id				セッションID
#			uri				セッション利用URI
#			$data{name} = value; or $data{name} = \%value
#			$data{name} = "" で削除
# 			id がない場合、新しいセッションを作成してデータを保存します。
# 			そのとき、変数 id に作成されたセッションID が設定されます。
# 				read_flag			保存と同時に読み込みを行います。
# 				%additional_values	古いデータを上書きせず追加する項目/$additional_values{項目名}=type
# 									ARRAY=配列に追加/+=数を加算/.=結合/,=カンマ区切りで結合
#
#	(7)-1 obj->save_by_name(id,uri,name,\%data,\%additional_values);
#		データを格納します。
#			name			データ名
#			%data が空のとき削除
#
# ---------------------------------------------------------------------------------------
#	(8)-1 obj->read_by_uri(id,uri,\%data,overwrite_disabled,\@name_order);
#		データを%dataに読み込みます。
#			overwrite_disabled		1=%data に既に値がある場合データの上書きを禁止する
#			name_order				name の格納順配列
#
#	(8)-2 obj->read_by_name(id,uri,name,\%data,overwrite_disabled);
#		データを%dataに読み込みます。
#
# ---------------------------------------------------------------------------------------
#	(9)-1 obj->reset_by_uri(id,uri);
#		uri のデータを削除します。
#
#	(9)-2 obj->reset_by_name(id,uri,@name);
#		name のデータを削除します。
#
# ---------------------------------------------------------------------------------------
#	(10) obj->log_out(id);
#		ユーザー情報を削除します。
#
######################################################################
#暗号化キー
$Session::salt = 'cM';

######################################################################
#オブジェクト作成
sub new ($$$$) {
    my ( $pkg, $store, $security, $lifetime ) = @_;
    $security = '' if !defined $security;
    $lifetime = 30 if !defined $lifetime || !$lifetime;
    my $this = {
        'store'    => $store,
        'security' => $security,
        'lifetime' => $lifetime
    };
    bless $this;
    $this;
}

######################################################################
#	(1) obj->is_available(id);
sub is_available ($$) {
    my ( $this, $id ) = @_;
    return ( $id && -f $this->{'store'} . '/' . $id ) ? 1 : 0;
}

######################################################################
#	(2) obj->get_user_info(id);
sub get_user_info ($$;$) {
    my ( $this, $sid, $want_email ) = @_;
    my ( $email, $pswd, $user_name, $remote_addr, $user_agent );
    local (*SESSION);
    if ( $sid && open( SESSION, $this->{'store'} . '/' . $sid ) ) {
        binmode SESSION, ":utf8";
        flock( SESSION, LOCK_SH );
        my $user_info = <SESSION>;
        close(SESSION);
        if ($user_info) {
            chomp $user_info;
            ( $email, $pswd, $remote_addr, $user_agent )
                = split( /\x09/, $user_info );
        }
        return $email if $want_email;
    }
    else {
        return ($want_email) ? undef : ( undef, undef );
    }

# 前回セッション発行時と違うセッションID or ユーザーIPが前回と異なるとき or ユーザーエージェントが前回と異なるとき => 認証情報を取得しない
    if ($this->is_invalid_session_id( $email, $sid )
        || (   $this->{'security'} =~ /ip/i
            && $remote_addr
            && ( !$ENV{'REMOTE_ADDR'} || $remote_addr ne $ENV{'REMOTE_ADDR'} )
        )
        || (   $this->{'security'} =~ /agent/i
            && $user_agent
            && (  !$ENV{'HTTP_USER_AGENT'}
                || $user_agent ne $ENV{'HTTP_USER_AGENT'} )
        )
        )
    {
        return ( undef, undef );
    }

    if ( $pswd && $pswd =~ /:/ ) {
        my %pass;
        while ( $pswd =~ /\G([0-9a-zA-Z_\.]+):([^;]*)(;|$)/g ) {
            $pass{$1} = $2;
        }
        return ( $email, \%pass );
    }
    else {
        return ( $email, $pswd );
    }
}

######################################################################
#	(3) obj->mk_session_id
sub mk_session_id ($) {
    my $this = shift;
    return $ENV{'UNIQUE_ID'}
        if defined $ENV{'UNIQUE_ID'} && $ENV{'UNIQUE_ID'};
    my ( $id, $salt );
    $id = time . $$ . $ENV{'REMOTE_ADDR'};
    $id = $cgi::Register->encrypt($id);
    $id =~ s/\$1\$//g;
    $id =~ s/\$//g;
    $id =~ s/\./\-/g;
    $id =~ s/\//\_/g;
    return $id;
}

######################################################################
#	(4) obj->set_user_info
sub set_user_info ($$$;$$) {
    my ( $this, $email, $pswd, $old_id, $logout ) = @_;
    my ($user_name,    $remote_addr, $user_agent,
        $new_session,  $old_session, $prev_id,
        $prev_session, $new_id,      $new_flag
    );
    $new_id      = $this->mk_session_id;
    $new_session = $this->{'store'} . '/' . $new_id;
    $prev_id     = $this->get_old_session_id($email) if $email;
    if ( !$old_id ) {
        $old_id = $prev_id;
        undef $prev_id;
    }
    $old_session  = $this->{'store'} . '/' . $old_id  if $old_id;
    $prev_session = $this->{'store'} . '/' . $prev_id if $prev_id;
    if ( ref $pswd ) {
        my $temp = '';
        while ( my ( $k, $v ) = each %$pswd ) {
            $k =~ s/[^0-9a-zA-Z_\.]/_/g;
            $v =~ s/;/\%3B/g;
            $temp .= ( ($temp) ? ';' : '' ) . $k . ':' . $v;
        }
        $pswd = $temp;
    }
    else {
        $pswd =~ s/[:;\x09\x0A\x0C\x0D\x20]//sg;
    }
    local ( *OLD, *PREV, *NEW );
    if ( open( NEW, ">>$new_session" ) ) {
        binmode NEW, ":utf8";
        flock( NEW, LOCK_EX );
        truncate( NEW, 0 );
        seek( NEW, 0, 0 );
        my $nodata = 1;
        my %done;
        print NEW ( ($logout)
            ? ''
            : ( ($email) ? $email : '' ) . "\x09"
                . $pswd . "\x09"
                . $ENV{'REMOTE_ADDR'} . "\x09"
                . $ENV{'HTTP_USER_AGENT'} )
            . "\x0A";
        if ( $old_session && open( OLD, $old_session ) ) {
            binmode OLD, ":utf8";
            flock( OLD, LOCK_SH );
            my $i = <OLD>;
            while ( $i = <OLD> ) {
                if ( $i && $i =~ /^([^\x09]+)/ ) {
                    $done{$1} = 1;
                    print NEW $i;
                }
            }
            close(OLD);
            unlink $old_session;
        }
        else {
            $new_flag = 1;
        }
        $nodata = 0 if %done;
        if ( $prev_session && open( PREV, $prev_session ) ) {
            binmode PREV, ":utf8";
            flock( PREV, LOCK_SH );
            my $i = <PREV>;
            while ( $i = <PREV> ) {
                if ( $i && $i =~ /^([^\x09]+)/ && !$done{$1} ) {
                    print NEW $i;
                    $nodata = 0;
                }
            }
            close(PREV);
            unlink $prev_session;
        }
        close(NEW);
        chmod( $env::file_mode, $new_session );
        $this->gc if defined $new_flag && $new_flag;
        $this->save_session_id( $email, $new_id );
        if ($logout) {
            $new_id = '';
            if ($nodata) {
                $this->save_session_id($email) if $email;
                unlink $new_session;
            }
        }
        return $new_id;
    }
    else {
        return $old_id;
    }
}

sub email_encrypt ($$) {
    my ( $this, $email ) = @_;
    my $encrypted;
    while ( $email =~ /([a-zA-Z0-9\/\.]{1,8})/g ) {
        $encrypted .= crypt( $1, $Session::salt );
    }
    $encrypted =~ s/\$1\$//g;
    $encrypted =~ s/\$//g;
    $encrypted =~ s/\./\-/g;
    $encrypted =~ s/\//\_/g;
    return $encrypted;
}

# アカウントとセッションIDの対応を保存
sub save_session_id ($$;$) {
    my ( $this, $email, $sid ) = @_;
    return 0 if !$email;
    my $path = $this->{'store'} . '/' . $this->email_encrypt($email);
    local (*SID);
    if ( !$sid ) {
        unlink $path;
        return 1;
    }
    elsif ( open( SID, ">>$path" ) ) {
        binmode SID, ":utf8";
        flock( SID, LOCK_EX );
        truncate( SID, 0 );
        seek( SID, 0, 0 );
        print SID $sid;
        close(SID);
        chmod( $env::file_mode, $path );
        return 1;
    }
    else {
        return 0;
    }
}

sub is_invalid_session_id ($$$) {
    my ( $this, $email, $sid ) = @_;
    return 0 if !$email;
    my $old_sid = $this->get_old_session_id($email);
    return ( $old_sid && $sid ne $old_sid ) ? 1 : 0;
}

sub get_old_session_id ($$) {
    my ( $this, $email ) = @_;
    return '' if !$email;
    my $path = $this->{'store'} . '/' . $this->email_encrypt($email);
    local (*SID);
    if ( open( SID, $path ) ) {
        my ($old_sid);
        binmode SID, ":utf8";
        flock( SID, LOCK_SH );
        $old_sid = <SID>;
        close(SID);
        return $old_sid;
    }
    else {
        return '';
    }
}

######################################################################
#	(5) obj->erase(id);
sub erase ($$) {
    my ( $this, $id ) = @_;
    return 1 if !defined $id;
    my $email = $this->get_user_info( $id, 1 );
    my $session = $this->{'store'} . '/' . $id;
    unlink $session;
    unlink $this->{'store'} . '/' . $this->email_encrypt($email) if $email;
    return ( -f $session ) ? 0 : 1;
}

######################################################################
#	(6) obj->gc;
sub gc ($) {
    my $this = shift;
    my ($id);
    local (*SESSION);
    opendir( SESSION, $this->{'store'} );
    while ( $id = readdir(SESSION) ) {
        my $session = $this->{'store'} . '/' . $id;
        unlink $session
            if $id !~ /^\./
            && -f $session
            && -M $session > $this->{'lifetime'};
    }
    closedir(SESSION);
    return $this;
}

######################################################################
#	(7)-1 obj->save_by_uri(id,uri,\%data,read_flag,\%additional_values);
sub save_by_uri ($$$$;$@) {
    my ( $this, $id, $uri, $hash, $read_flag, $additional_values ) = @_;
    my ( $session, $session_temp, $no_id_flag, $has_data, %done );
    if ( !defined $id || !$id || $id =~ /^\.\.\// || $id =~ /pswd=/ ) {
        $no_id_flag   = 1;
        $id           = $this->mk_session_id;
        $_[1]         = $id;
        $session_temp = $this->{'store'} . '/' . $id;
    }
    else {
        $session      = $this->{'store'} . '/' . $id;
        $session_temp = $this->{'store'} . '/' . $id . '.temp';
    }
    local ( *SESSION, *TEMP );
    if ( open( TEMP, ">>$session_temp" ) ) {
        binmode TEMP, ":utf8";
        flock( TEMP, LOCK_EX );
        truncate( TEMP, 0 );
        seek( TEMP, 0, 0 );
        if ( defined $no_id_flag && $no_id_flag ) {
            print TEMP "\x0A";
        }
        elsif ( open( SESSION, $session ) ) {
            binmode SESSION, ":utf8";
            flock( SESSION, LOCK_SH );
            my $data = <SESSION>;
            print TEMP $data;
            $has_data = 1 if $data =~ /\S/;
            while ( $data = <SESSION> ) {
                if ( $data =~ /^$uri\x09([^\x09]+)\x09(.*)\x0A/ ) {
                    my ( $key, $value ) = ( $1, $2 );
                    if ( defined $hash->{$key} ) {
                        $done{$key} = 1;
                        next if !ref $hash->{$key} && $hash->{$key} eq '';
                        $this->add( $hash->{$key}, $value,
                            $additional_values )
                            if ref $hash->{$key} && ref $additional_values;
                        $data = $this->data( $uri, $key, $hash->{$key} );
                    }
                    elsif ($read_flag) {
                        $done{$key} = 1;
                        $this->read_data( $hash, $key, $value );
                    }
                }
                if ( $data =~ /\S/ ) {
                    print TEMP $data;
                    $has_data = 1;
                }
            }
            close(SESSION);
        }
        else {
            close(TEMP);
            unlink $session_temp;
            return '';
        }
        while ( my ( $k, $v ) = each %$hash ) {
            if ( !$done{$k} && defined $v && ( ref $v || $v ne '' ) ) {
                my $data = $this->data( $uri, $k, $v );
                if ( $data =~ /\S/ ) {
                    print TEMP $data;
                    $has_data = 1;
                }
            }
        }
        close(TEMP);
        if ( defined $no_id_flag && $no_id_flag ) {
            chmod( $env::file_mode, $session_temp );
            unlink $session_temp if !$has_data;
        }
        else {
            rename( $session_temp, $session );
            chmod( $env::file_mode, $session );
            unlink $session if !$has_data;
        }
        return $id;
    }
    else {
        return '';
    }
}

#	(7)-2 obj->save_by_name(id,uri,name,\%data,\%additional_values);
sub save_by_name ($$$$;$@) {
    my ( $this, $id, $uri, $name, $hash, $additional_values ) = @_;
    my ( $session, $session_temp, $no_id_flag, $done, $has_data );
    if ( !defined $id || !$id || $id =~ /^\.\.\// || $id =~ /pswd=/ ) {
        $no_id_flag   = 1;
        $id           = $this->mk_session_id;
        $_[1]         = $id;
        $session_temp = $this->{'store'} . '/' . $id;
    }
    else {
        $session      = $this->{'store'} . '/' . $id;
        $session_temp = $this->{'store'} . '/' . $id . '.temp';
    }
    local ( *SESSION, *TEMP );
    if ( open( TEMP, ">>$session_temp" ) ) {
        binmode TEMP, ":utf8";
        flock( TEMP, LOCK_EX );
        truncate( TEMP, 0 );
        seek( TEMP, 0, 0 );
        if ( defined $no_id_flag && $no_id_flag ) {
            print TEMP "\x0A";
        }
        elsif ( open( SESSION, $session ) ) {
            binmode SESSION, ":utf8";
            flock( SESSION, LOCK_SH );
            my $data = <SESSION>;
            print TEMP $data;
            $has_data = 1 if $data =~ /\S/;
            while ( $data = <SESSION> ) {
                if ( $data =~ /^$uri\x09$name\x09(.*)\x0A/ ) {
                    my $value = $1;
                    if ( defined $hash && ( ref $hash || $hash ne '' ) ) {
                        $done = 1;
                        $this->add( $hash, $value, $additional_values )
                            if ref $hash && ref $additional_values;
                        $data = $this->data( $uri, $name, $hash );
                    }
                    else {
                        next;
                    }
                }
                if ( $data =~ /\S/ ) {
                    print TEMP $data;
                    $has_data = 1;
                }
            }
            close(SESSION);
        }
        else {
            close(TEMP);
            unlink $session_temp;
            return '';
        }
        if ( !$done && defined $hash && ( ref $hash || $hash ne '' ) ) {
            my $data = $this->data( $uri, $name, $hash );
            if ( $data =~ /\S/ ) {
                print TEMP $data;
                $has_data = 1;
            }
        }
        close(TEMP);
        if ( defined $no_id_flag && $no_id_flag ) {
            chmod( $env::file_mode, $session_temp );
            unlink $session_temp if !$has_data;
        }
        else {
            rename( $session_temp, $session );
            chmod( $env::file_mode, $session );
            unlink $session if !$has_data;
        }
        return $id;
    }
    else {
        return '';
    }
}

sub data ($$$$;$) {
    my ( $this, $uri, $name, $data, $nokey ) = @_;
    return '' if !defined $data || $data eq "\x7F";
    my ($out);
    $out = $uri . "\x09" . $name . "\x09" if !$nokey;
    if ( ref $data eq 'HASH' ) {
        my ( $k, $v, $value );
        while ( ( $k, $v ) = each %$data ) {
            $value .= $this->url_encode( $k, $v );
        }
        if ( defined $value ) {
            $value =~ s/^\&//;
            $out .= $value;
        }
    }
    else {
        $out .= '=' . $data;
    }
    $out .= "\x0A";
    return $out;
}

sub url_encode ($$$) {
    my ( $this, $key, $value ) = @_;
    my $data = '';
    if ( ref $value eq 'ARRAY' ) {
        my $n = 0;
        foreach my $v (@$value) {
            $data .= $this->url_encode( $key . '^' . $n++, $v );
        }
    }
    elsif ( ref $value eq 'HASH' ) {
        while ( my ( $k, $v ) = each %$value ) {
            $data .= $this->url_encode( $key . ':' . $k, $v );
        }
    }
    elsif ( defined $key ) {
        $value = $$value if ref $value eq 'SCALAR';
        utf8::encode($key)   if utf8::is_utf8($key);
        utf8::encode($value) if utf8::is_utf8($value);
        $key
            =~ s/([\x00-\x1F\x21-\x29\x2B\x2C\x2F\x3A-\x3F\x5B-\x5E\x60\x7B-\xFF])/'%' . sprintf("%.2X", unpack("C", $1))/eg;
        $value
            =~ s/([\x00-\x1F\x21-\x29\x2B\x2C\x2F\x3A-\x3F\x5B-\x5E\x60\x7B-\xFF])/'%' . sprintf("%.2X", unpack("C", $1))/eg
            if defined $value;
        $data .= '&' . $key . '=' . ( ( defined $value ) ? $value : '' );
    }
    return $data;
}

sub url_decode ($$) {
    my ( $this, $value ) = @_;
    utf8::encode($value) if utf8::is_utf8($value);
    $value =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C",hex($1))/ge;
    utf8::decode($value) if !utf8::is_utf8($value);
    return $value;
}

sub add ($$$$) {
    my ( $this, $hash, $value, $additional_values ) = @_;
    if ( !ref $hash && defined $value && $value =~ s/^=// ) {
        $hash .= ( ($hash) ? '&' : '' ) . $value
            if $additional_values
            && !ref $additional_values
            && $additional_values eq '=';
    }
    else {
        while ( $value =~ /\G([^=\&]*)=([^\&]*)\&?/sgo ) {
            my ( $k, $v ) = ( $1, $2 );
            $k = $this->url_decode($k);
            if ( ref $additional_values && $additional_values->{$k} ) {
                $k = $this->url_decode($v);
                if ( !defined $hash || !defined $hash->{$k} ) {
                    $hash->{$k} = $v;
                }
                elsif ( ref $hash && ref $hash->{$k} eq 'ARRAY' ) {
                    push( @{ $hash->{$k} }, $v );
                }
                elsif ( $hash->{$k} && !ref $hash->{$k} ) {
                    if ( $additional_values->{$k} eq '+' ) {
                        $hash->{$k} += $v;
                    }
                    elsif ( $additional_values->{$k} eq '.' ) {
                        $hash->{$k} = $v . $hash->{$k};
                    }
                    elsif ( $additional_values->{$k} eq ',' ) {
                        $hash->{$k} = $v . ',' . $hash->{$k};
                    }
                    else {
                        my $temp = $hash->{$k};
                        $hash->{$k} = [$temp];
                    }
                }
            }
        }
    }
    return $this;
}

######################################################################
#	(8) obj->read_by_uri(id,uri,\%data,overwrite_disabled);
sub read_by_uri ($$$$;$$) {
    my ( $this, $id, $uri, $hash, $overwrite_disabled, $name_order ) = @_;
    if ( !$id || $id =~ /^\.\.\// || $id =~ /pswd=/ ) {
        $_[1] = '';
        return 0;
    }
    my $session = $this->{'store'} . '/' . $id;
    local (*SESSION);
    if ( open( SESSION, $session ) ) {
        binmode SESSION, ":utf8";
        flock( SESSION, LOCK_SH );
        my $data = <SESSION>;
        while ( $data = <SESSION> ) {
            if ( $data =~ /^$uri\x09([^\x09]+)\x09(.*)\x0A/ ) {
                push( @$name_order, $1 )
                    if ref $name_order && ref $name_order eq 'ARRAY';
                $this->read_data( \$hash->{$1}, $2, $overwrite_disabled );
            }
        }
        close(SESSION);
        return 1;
    }
    else {
        $_[1] = '';
        return 0;
    }
}

sub read_by_name ($$$$$;$) {
    my ( $this, $id, $uri, $name, $hash, $overwrite_disabled ) = @_;
    if ( !$id || $id =~ /^\.\.\// || $id =~ /pswd=/ ) {
        $_[1] = '';
        return 0;
    }
    my $session = $this->{'store'} . '/' . $id;
    local (*SESSION);
    if ( open( SESSION, $session ) ) {
        binmode SESSION, ":utf8";
        flock( SESSION, LOCK_SH );
        my $data = <SESSION>;
        while ( $data = <SESSION> ) {
            if ( $data =~ /^$uri\x09$name\x09(.*)\x0A/ ) {
                $this->read_data( $hash, $2, $overwrite_disabled );
                last;
            }
        }
        close(SESSION);
        return 1;
    }
    else {
        $_[1] = '';
        return 0;
    }
}

sub read_data ($$$;$) {
    my ( $this, $hash, $value, $overwrite_disabled ) = @_;
    chomp $value;
    if ( $value =~ s/^=// && ref $hash eq 'SCALAR' ) {
        utf8::encode($value) if utf8::is_utf8($value);
        $$hash = $value;
    }
    else {
        $hash = \%{$$hash} if ref $hash eq 'SCALAR';
        while ( $value =~ /\G([^=\&]*)=([^\&]*)[\&]?/sgo ) {
            my ( $k, $v ) = ( $this->url_decode($1), $this->url_decode($2) );
            my $temp = \$hash;
            while ( $k =~ /\G([^:\^]+)([\^:]|$)/sg ) {
                $temp
                    = ( ref $$temp eq 'ARRAY' )
                    ? \$$temp->[$1]
                    : \$$temp->{$1};
                if ( !$2 ) {
                    last;
                }
                elsif ( !ref $$temp ) {
                    if ( $2 eq '^' ) {
                        $$temp = [];
                    }
                    else {
                        $$temp = {};
                    }
                }
            }
            if ( $overwrite_disabled && defined $$temp ) {
                next;
            }
            elsif ( defined $$temp && !ref $$temp && $$temp ne '' ) {
                $$temp = [ $$temp, $v ];
            }
            elsif ( ref $$temp eq 'ARRAY' ) {
                push( @$$temp, $v );
            }
            else {
                $$temp = $v;
            }
        }
    }
}

######################################################################
#	(9) obj->reset_by_uri(id,@uri);
sub reset_by_uri ($$@) {
    my ( $this, $id, @uri ) = @_;
    return 1 if !defined $id || !$id;
    my ( $session, $session_temp, $has_data );
    $session      = $this->{'store'} . '/' . $id;
    $session_temp = $this->{'store'} . '/' . $id . '.temp';
    local ( *SESSION, *TEMP );
    if ( open( SESSION, $session ) ) {
        binmode SESSION, ":utf8";
        flock( SESSION, LOCK_SH );
        if ( open( TEMP, ">>$session_temp" ) ) {
            binmode TEMP, ":utf8";
            flock( TEMP, LOCK_EX );
            truncate( TEMP, 0 );
            seek( TEMP, 0, 0 );
            my $data = <SESSION>;
            $has_data = 1 if $data =~ /\S/;
            print TEMP $data;
            while ( $data = <SESSION> ) {

                if ( @uri
                    && grep( defined $_ && $_ ne '' && $data =~ /^$_\x09/,
                        @uri ) )
                {
                    next;
                }
                elsif ( $data =~ /\S/ ) {
                    print TEMP $data;
                    $has_data = 1;
                }
            }
            close(TEMP);
        }
        else {
            close(SESSION);
            return 0;
        }
        close(SESSION);
        rename( $session_temp, $session );
        chmod( $env::file_mode, $session );
        unlink $session if !$has_data;
        return 1;
    }
    else {
        return 0;
    }
}

#	(9) obj->reset_by_name(id,uri,name);
sub reset_by_name ($$$@) {
    my ( $this, $id, $uri, @name ) = @_;
    return 1 if !defined $id || !$id;
    my ( $session, $session_temp, $has_data );
    $session      = $this->{'store'} . '/' . $id;
    $session_temp = $this->{'store'} . '/' . $id . '.temp';
    local ( *SESSION, *TEMP );
    if ( open( SESSION, $session ) ) {
        binmode SESSION, ":utf8";
        flock( SESSION, LOCK_SH );
        if ( open( TEMP, ">>$session_temp" ) ) {
            binmode TEMP, ":utf8";
            flock( TEMP, LOCK_EX );
            truncate( TEMP, 0 );
            seek( TEMP, 0, 0 );
            my $data = <SESSION>;
            $has_data = 1 if $data =~ /\S/;
            print TEMP $data;
            while ( $data = <SESSION> ) {

                if (grep( defined $_
                            && $_ ne ''
                            && $data =~ /^$uri\x09$_\x09/,
                        @name )
                    )
                {
                    next;
                }
                elsif ( $data =~ /\S/ ) {
                    print TEMP $data;
                    $has_data = 1;
                }
            }
            close(TEMP);
        }
        else {
            close(SESSION);
            return 0;
        }
        close(SESSION);
        rename( $session_temp, $session );
        chmod( $env::file_mode, $session );
        unlink $session if !$has_data;
        return 1;
    }
    else {
        return 0;
    }
}

######################################################################
#	(10) obj->log_out(id);
sub log_out ($$$) {
    my ( $this, $id ) = @_;
    return ( !$id )
        ? ''
        : $this->set_user_info( $this->get_user_info( $id, 1 ), '', $id, 1 );
}

1;
